from redis import Redis
from datetime import datetime, timedelta
import json
from dateutil.parser import parse as dtparse
from EVECelery.exceptions.utils import ErrorLimitExceeded
import socket
import os


class ErrorLimiter(object):
    @classmethod
    def max_remaining_errors(cls) -> int:
        """Return the remaining error limit threshold.
        If the remaining error count is less than or equal to this value then ErrorLimitExceeded exceptions are raised.

        :return: Errors remaining allowed before raising ErrorLimitExceeded exceptions.
        """
        return 0

    @classmethod
    def default_error_limit_remain(cls) -> int:
        """Returns the default error limit remaining if the Redis key is created before being set by headers.

        :return: The default error limit remaining
        """
        raise NotImplementedError

    @classmethod
    def default_error_limit_reset(cls) -> int:
        """Returns the default reset internal in seconds if the Redis key is created before being set by headers.

        :return: The default reset interval in seconds.
        """
        raise NotImplementedError

    @classmethod
    def get_key(cls) -> str:
        """Returns the Redis key name for storing error information.

        :return: Redis key name
        :rtype: str
        """
        raise NotImplementedError

    @classmethod
    def get_lock_key(cls) -> str:
        """Returns the lock Redis key name for storing lock information.

        :return: Redis key name
        :rtype: str
        """
        return f"Lock-{cls.get_key()}"

    @classmethod
    def get_redis(cls, redis: Redis):
        """Return the data from Redis for the ErrorLimit key.

        :param redis: The redis client
        :return: Error limit information or None if not set.
        :rtype: dict or None
        """
        cached_data = redis.get(cls.get_key())
        if cached_data:
            return json.loads(cached_data)
        else:
            return None

    @classmethod
    def _create_default_limit(cls, redis: Redis):
        """Set the default error info if the key does not exist in Redis.
        This class does not establish a lock.
        This function should only be used by calling functions with the lock obtained.

        :param redis: The redis client
        :return: None
        """
        d = {"error_limit_remain": cls.default_error_limit_remain(),
             "error_limit_reset": cls.default_error_limit_reset(),
             "updated_at": str(datetime.utcnow())}
        redis.set(name=cls.get_key(), value=json.dumps(d), ex=cls.default_error_limit_reset())

    @classmethod
    def check_limit(cls, redis: Redis):
        """Ensure the remaining error limit is not breached.

        :param redis: The redis client
        :return: None
        :raises EVECelery.exceptions.utils.ErrorLimitExceeded: If the remaining error limit is below the allowed threshold.
        """
        with redis.lock(cls.get_lock_key(), blocking_timeout=5, timeout=300):
            current_error_info = cls.get_redis(redis)
            if not current_error_info:  # error key is not set so no errors
                return
            else:
                if current_error_info.get("error_limit_remain") <= cls.max_remaining_errors():
                    raise ErrorLimitExceeded

    @classmethod
    def update_limit(cls, redis: Redis, error_limit_remain: int, error_limit_reset: int,
                     time: datetime):
        """Updates the remaining error limit.

        :param redis: The redis client
        :param error_limit_remain: The error limit remaining.
        :param error_limit_reset: The error limit reset in seconds.
        :param time: The date time from the return header.
        :return: None
        """
        with redis.lock(cls.get_lock_key(), blocking_timeout=5, timeout=300):
            current_error_info = cls.get_redis(redis)
            if current_error_info:  # key exits
                if dtparse(current_error_info["updated_at"]) > time:  # incoming request is outdated
                    return
            else:  # key does not exit in Redis
                if (datetime.utcnow() - timedelta(seconds=10)) > time:  # request is older than 10 seconds so ignore
                    return
            d = {"error_limit_remain": error_limit_remain, "error_limit_reset": error_limit_reset,
                 "updated_at": str(time)}
            redis.set(name=cls.get_key(), value=json.dumps(d), ex=error_limit_reset)

    @classmethod
    def decrement_limit(cls, redis: Redis, time: datetime):
        """Decrement the remaining error limit by one. Useful in timeout or missing header situations.

        :param redis: The redis client
        :param time: The date time from the calling function.
        :return: None
        """
        with redis.lock(cls.get_lock_key(), blocking_timeout=5, timeout=300):
            current_error_info = cls.get_redis(redis)
            if current_error_info:  # key exits
                if dtparse(current_error_info["updated_at"]) > time:  # incoming request is outdated
                    return
            else:  # key does not exit in Redis
                if (datetime.utcnow() - timedelta(seconds=10)) > time:  # request is older than 10 seconds so ignore
                    return
                else:  # we need to init the new default
                    cls._create_default_limit(redis)
                    current_error_info = cls.get_redis(redis)
            new_reset_seconds = max(redis.ttl(cls.get_key()), 1)
            d = {"error_limit_remain": max(current_error_info["error_limit_remain"] - 1, 0),
                 "error_limit_reset": new_reset_seconds,
                 "updated_at": str(time)}
            redis.set(name=cls.get_key(), value=json.dumps(d), ex=new_reset_seconds)


class ESIErrorLimiter(ErrorLimiter):
    @classmethod
    def default_error_limit_remain(cls):
        return 100

    @classmethod
    def default_error_limit_reset(cls) -> int:
        return 60

    @classmethod
    def get_key(cls):
        return f"ESIErrorLimiter-{socket.gethostname()}"

    @classmethod
    def max_remaining_errors(cls) -> int:
        return int(os.environ.get('EVECelery_MaxConcurrency', 4))



from EVECelery.tasks.BaseTasks.ESIResqust import ESIRequest
from EVECelery.exceptions.tasks import InputValidationError
from .CategoryInfo import CategoryInfo


class GroupInfo(ESIRequest):
    def get_key(self, group_id: int):
        return f"GroupInfo-{group_id}"

    def route(self, group_id: int):
        return f"/universe/groups/{group_id}"

    def _hook_after_esi_success(self, esi_response: dict) -> None:
        category_id = esi_response.get("category_id")
        if category_id:
            CategoryInfo().get_async(ignore_result=True, category_id=category_id)

    def validate_inputs(self, group_id: int) -> None:
        try:
            int(group_id)
        except ValueError:
            raise InputValidationError("Input parameter must be an integer.")





from EVECelery.tasks.BaseTasks.ESIResqust import ESIRequest
from EVECelery.exceptions.tasks import InputValidationError
from .GroupInfo import GroupInfo


class TypeInfo(ESIRequest):
    def get_key(self, type_id: int):
        return f"TypeInfo-{type_id}"

    def route(self, type_id: int):
        return f"/universe/types/{type_id}"

    def _hook_after_esi_success(self, esi_response: dict) -> None:
        group_id = esi_response.get("group_id")
        if group_id:
            GroupInfo().get_async(ignore_result=True, group_id=group_id)

    def validate_inputs(self, type_id: int) -> None:
        try:
            int(type_id)
        except ValueError:
            raise InputValidationError("Input parameter must be an integer.")




from EVECelery.tasks.BaseTasks.ESIResqust import ESIRequest
from EVECelery.exceptions.tasks import InputValidationError


class Route(ESIRequest):
    def get_key(self, origin: int, destination: int):
        return f"Route-{origin}-{destination}"

    def route(self, origin: int, destination: int):
        return f"/route/{origin}/{destination}"

    def validate_inputs(self, origin: int, destination: int):
        try:
            int(origin)
            int(destination)
        except ValueError:
            raise InputValidationError("Input parameter must be an integer.")




from EVECelery.tasks.BaseTasks.ESIResqust import ESIRequest


class PricesList(ESIRequest):
    def get_key(self):
        return f"PricesList"

    def route(self):
        return f"/markets/prices"

    def validate_inputs(self, **kwargs) -> None:
        return





from EVECelery.exceptions.tasks import NotResolved, InputValidationError
from EVECelery.exceptions.utils import ErrorLimitExceeded
from EVECelery.utils.ErrorLimiter import ESIErrorLimiter
from EVECelery.utils.RequestHeaders import RequestHeaders
from redis import Redis
from celery.result import AsyncResult
import json
import requests
from datetime import datetime
from dateutil.parser import parse as dtparse
from .BaseTask import BaseTask


class ESIRequest(BaseTask):
    autoretry_for = (Exception,)
    max_retries = 3
    retry_backoff = 5
    retry_backoff_max = 600
    retry_jitter = False

    def ttl_404(self) -> int:
        """Returns the redis TTL for caching ESI responses that errored with a 404 - not found

        :return: Seconds to cache a 404 not found ESI response in Redis
        :rtype: int
        """
        return 86400

    def get_key(self, **kwargs) -> str:
        """Returns the Redis key name for storing ESI responses

        :param kwargs: ESI request parameters
        :return: Redis key for storing the value of the ESI response
        :rtype: str
        """
        raise NotImplementedError

    def get_lock_key(self, **kwargs) -> str:
        """Returns the Redis key name for storing ESI locks

        :param kwargs: ESI request parameters
        :return: Redis key for storing ESI locks
        :rtype: str
        """
        k = self.get_key(**kwargs)
        return f"Lock-{k}"

    def base_url(self) -> str:
        """Base URL for ESI requests

        :return: ESI base request URL
        :rtype: str
        """
        return "https://esi.evetech.net/latest"

    def route(self, **kwargs) -> str:
        """ESI route with input request parameters

        :param kwargs: ESI request parameters to fill in the ESI request string
        :return: ESI route with request parameters
        :rtype: str
        """
        raise NotImplementedError

    def request_url(self, **kwargs) -> str:
        """ESI request URL with request parameters

        :param kwargs: ESI request parameters to fill in the ESI request string
        :return: ESI request URL with request parameters
        :rtype: str
        """
        return f"{self.base_url()}{self.route(**kwargs)}/?datasource=tranquility"

    def get_cached(self, **kwargs):
        """Get the cached response from ESI immediately without invoking an ESI call.

        :param kwargs: ESI request parameters
        :return: Dictionary or list containing cached response from ESI.
            If ESI returned a 404 error dictionary the response will be in the form
            {"error": error_message, "error_code": 404}
            Only /universe/factions/ and /markets/prices/ returns a list, all else return dictionaries.
        :rtype: dict or list
        :raises EVECelery.exceptions.ESI.NotResolved: If the request has not yet been resolved by ESI.
        :raises EVECelery.exceptions.ESI.InputValidationError: If an input ESI parameter contains
            invalid syntax or is a known invalid ID
        """
        self.validate_inputs(**kwargs)
        cached_data = self.redis_cache.get(self.get_key(**kwargs))
        if cached_data:
            return json.loads(cached_data)
        else:
            raise NotResolved(f"Not resolved. ESI request parameters: {kwargs}")

    def get_async(self, ignore_result: bool = False, **kwargs) -> AsyncResult:
        """Returns an async result / promise representing the future evaluation of a task.
        This function does not block the calling process and returns immediately.

        :param ignore_result: Set false to store result in celery result backend, else ignore the result.
        :type ignore_result: bool
        :param kwargs: ESI request parameters
        :return: Promise for a future evaluation of a celery task
        :rtype: celery.result.AsyncResult
        """
        self.validate_inputs(**kwargs)
        return self.apply_async(ignore_result=ignore_result, kwargs=kwargs)

    def get_sync(self, timeout: float = 10, **kwargs):
        """Call a task and block until the result is set.

        :param timeout: The time in seconds to block waiting for the task to complete.
        Setting this to None blocks forever.
        :type timeout: float
        :param kwargs: ESI request parameters
        :return: Dictionary or list containing response from ESI.
            If ESI returned a 404 dictionary error the response will be in the form
            {"error": error_message, "error_code": 404}
            Only /universe/factions/ and /markets/prices/ returns a list, all else return dictionaries.
        :rtype: dict or list
        """
        return self.get_async(ignore_result=False, **kwargs).get(timeout=timeout, propagate=True)

    def run(self, *args, **kwargs):
        return self._request_esi(**kwargs)

    def _request_esi(self, **kwargs):
        """Gets the ESI cached response.
        If the response is not yet cached or hasn't been resolved then perform an ESI call caching the new response.

        This function should not be called outside of celery tasks and should only be invoked
        by the task function handling a lookup queue.

        :param redis: The redis client
        :param kwargs: ESI request parameters
        :return: Dictionary containing response from ESI.
            If ESI returned a 404 error the response will be in the form
            {"error": error_message, "error_code": 404}
            If the response doesn't require request inputs then list is usually returned (list factions, prices, etc).
            If the response requires inputs a dictionary is usually returned.
            Only /universe/factions/ and /markets/prices/ returns a list, all else return dictionaries.
        :rtype: dict or list
        :raises EVECelery.exceptions.utils.ErrorLimitExceeded: If the remaining error limit is below the allowed threshold.
        """
        lookup_key = self.get_key(**kwargs)
        lock_key = self.get_lock_key(**kwargs)
        with self.redis_cache.lock(lock_key, blocking_timeout=15, timeout=300):
            try:
                return self.get_cached(**kwargs)
            except NotResolved:
                pass
            ESIErrorLimiter.check_limit(self.redis_cache)
            rheaders = {}
            try:
                resp = requests.get(self.request_url(**kwargs), headers=RequestHeaders.get_headers(),
                                    timeout=5, verify=True)
                rheaders = resp.headers
                if resp.status_code == 200:
                    d = resp.json()
                    ttl_expire = int(max(
                        (dtparse(rheaders["expires"], ignoretz=True) - datetime.utcnow()).total_seconds(),
                        1)
                    )
                    self.redis_cache.set(name=lookup_key, value=json.dumps(d), ex=ttl_expire)
                    self._hook_after_esi_success(d)
                    ESIErrorLimiter.update_limit(self.redis_cache,
                                                 error_limit_remain=int(rheaders["x-esi-error-limit-remain"]),
                                                 error_limit_reset=int(rheaders["x-esi-error-limit-reset"]),
                                                 time=dtparse(rheaders["date"], ignoretz=True)
                                                 )
                    return json.loads(self.redis_cache.get(lookup_key))
                elif resp.status_code == 400 or resp.status_code == 404:
                    d = {"error": str(resp.json().get("error")), "error_code": resp.status_code}
                    self.redis_cache.set(name=lookup_key, value=json.dumps(d), ex=self.ttl_404())
                    ESIErrorLimiter.update_limit(self.redis_cache,
                                                 error_limit_remain=int(rheaders["x-esi-error-limit-remain"]),
                                                 error_limit_reset=int(rheaders["x-esi-error-limit-reset"]),
                                                 time=dtparse(rheaders["date"], ignoretz=True)
                                                 )
                    return json.loads(self.redis_cache.get(lookup_key))
                else:
                    resp.raise_for_status()
            except Exception as ex:
                try:
                    ESIErrorLimiter.update_limit(self.redis_cache,
                                                 error_limit_remain=int(rheaders["x-esi-error-limit-remain"]),
                                                 error_limit_reset=int(rheaders["x-esi-error-limit-reset"]),
                                                 time=dtparse(rheaders["date"], ignoretz=True)
                                                 )
                except KeyError:
                    ESIErrorLimiter.decrement_limit(self.redis_cache, datetime.utcnow())
                raise ex

    def _hook_after_esi_success(self, esi_response) -> None:
        """Code to run with esi_response data after there was a successful 200 response from ESI.
        For example: use this function to optionally queue up additional ESI calls for ids returned.

        :param esi_response: ESI response from an API call
        :type esi_response: dict or list
        :rtype: None
        """
        return

    def validate_inputs(self, **kwargs) -> None:
        """Run validation checks before submitting an ESI request or hitting the cache.

        :param kwargs: ESI request parameters
        :return: None
        :raises EVECelery.exceptions.ESI.InputValidationError: If an input ESI parameter contains
            invalid syntax or is a known invalid ID
        """
        raise NotImplementedError



__author_email__ = 'maintainers@eveinsight.net'
__description__ = 'Task queue framework for building tools that interact with the EVE Online ESI API using Celery, RabbitMQ, and Redis.'
__license__ = 'MIT License'
__title__ = 'EVECelery'
__url__ = 'https://github.com/NatEVETools/EVECelery'
__version__ = '0.17'



#!/usr/bin/env python

import setuptools

pkg_info = {}
with open("EVECelery/__version__.py", "r", encoding="utf-8") as f:
    for l in f.readlines():
        data = l.split('=')
        pkg_info[data[0].strip()] = (data[1].replace("'", "")).strip()

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(name='EVECelery',
                 version=pkg_info['__version__'],
                 description=pkg_info['__description__'],
                 license=pkg_info['__license__'],
                 long_description=long_description,
                 long_description_content_type="text/markdown",
                 author_email=pkg_info['__author_email__'],
                 url=pkg_info['__url__'],
                 packages=setuptools.find_packages(include=["EVECelery", "EVECelery.*"]),
                 install_requires=[
                     'Celery~=5.2',
                     'pika~=1.3',
                     'python-dateutil~=2.8.2',
                     'redis~=4.1',
                     'requests~=2.27',
                     'pydantic~=1.10'
                 ],
                 python_requires=">=3.7",
                 entry_points={
                     'console_scripts': [
                         'eve-celery = EVECelery:main'
                     ]
                 }
                 )



from tests.TestUtils import *
import pytest
from EVECelery.CeleryApps import CeleryWorker
from redis.exceptions import ConnectionError
from redis.exceptions import AuthenticationError


class TestCeleryWorker:
    def test_init_no_running_servers(self, mock_env_no_servers):
        """test responses on init when no servers are running"""
        with pytest.raises(ConnectionError):
            CeleryWorker(connection_check=True)

        CeleryWorker()  # should not raise an exception since connection_check is by default false

    def test_init_running_servers(self, mock_env_rabbitmq_redis):
        """test responses on init when servers are running"""
        CeleryWorker(connection_check=True)  # should not raise an exception

    def test_init_running_servers_redis_bad_auth(self, mock_env_rabbitmq_redis, monkeypatch):
        """test bad auth to redis"""
        monkeypatch.setenv('EVECelery_Redis_ResultBackend_Password', 'WrongPass')
        with pytest.raises(AuthenticationError):
            CeleryWorker(connection_check=True)  # should raise auth error



from EVECelery.CeleryApps import CeleryWorker, CeleryBeat
import sys


def main():
    if "--beat" not in sys.argv:
        celery_server = CeleryWorker
    else:
        celery_server = CeleryBeat

    c = celery_server(connection_check=True)
    c.start()


if __name__ == '__main__':
    main()
